.tabs, .tabs2, .tabs3, .navpath ul {
    background-image: none;
    background-color: #333;
    border: none;
    border-bottom: 1px solid #575757;
}

.tablist li, .navpath li {
    background-image: none;
    background-color: #333;
}

.tablist a, .navpath li.navelem a {
    color: #ccc;
    text-shadow: 0px 1px 1px black;
}

.tablist a:hover, .navpath li.navelem a:hover {
    background-image: none;
    background-color: #444;
    color: #ccc;
}
